#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
import re
import subprocess
import json

TVE_REBOOT_DIR = '/config/_reboot/tve/now'
SSH_PORT_FILE = '/config/ssh_port'

def request_tve_reboot():
    # Creation of TVE_REBOOT_DIR directory triggers the restart.
    # If the dir is present then remove it first, then create it. Otherwise the trigger will not sense the creation
    if os.path.exists(TVE_REBOOT_DIR):
        os.rmdir(TVE_REBOOT_DIR)

    # Create dirs
    os.makedirs(TVE_REBOOT_DIR)

def is_svc_masked(svc):
    try:
        state = subprocess.check_output(['systemctl', 'is-enabled', svc])
        if state.decode().rstrip() == 'masked':
            return True
    except subprocess.CalledProcessError as e:
        if e.output.decode().rstrip() == 'masked':
            return True

    return False

def get_hw_type():
    if not os.path.isfile('/config/.hwtype'):
        return 'intel'
    with open('/config/.hwtype', 'r') as hw_type_file:
        return hw_type_file.readline().splitlines()[0]

def shell_return_ok(self, cmd):
    try:
        self.execute_shell_command(cmd)
    except Exception as ex:
        self.logger.error(ex)
        return False
    return True

def get_ifname(self, name):
    ifname = name

    if name.find("data") > -1:
        ifname = "br-{}".format(name)

    if 'lte' in name:
        ifname = "br-{}".format(name)

    if 'wifi' in name:
        ifname = "br-{}".format(name)

    return ifname

def set_pci_whitelist(self, action, devname, physical_network, current_list):
    import ast
    import json
    try:
        current_list = ast.literal_eval(str(current_list).strip())
    except Exception as ex:
        self.logger.error("Error parsing PCI current list '{}'".format(current_list))
        self.logger.error(ex)
        current_list = []
    if action == 'add':
        new_pci = {"devname": devname, "physical_network": physical_network}
        current_list.append(new_pci)
    elif action == 'remove':
        current_list[:] = [pci for pci in current_list if pci['devname'] != devname]
    return json.dumps(current_list)

def init_pci_data(self):
    self.pciids = dict()
    self.igb_vfs = dict()
    self.initramfs_changed = False
    self.igb_recalc_needed = False
    lines = ""

    if os.path.exists("/etc/udev/rules.d/70-persistent-net.rules"):
        with open("/etc/udev/rules.d/70-persistent-net.rules", "r") as f:
            lines = f.readlines()

    for line in lines:
        if 0 == len(line.strip()):
            continue
        line = line.replace('==', '=')
        line = line.replace(' ', '')
        rules = line.strip().split(',')
        rules_dict = dict()
        for rule in rules:
            (key, val) = tuple(rule.split('='))
            rules_dict[key] = val.replace('\"', '')

        if 'net' == rules_dict.get('SUBSYSTEM', 'none') and 'none' != rules_dict.get('NAME', 'none'):
            name = rules_dict.get('NAME', 'none')
            pci = rules_dict.get('KERNELS', 'none')
            tempdict = {name: pci}
            tempdict.update(self.pciids)
            self.pciids = tempdict

    for name, pci in self.pciids.items():
        if os.path.exists("/sys/bus/pci/drivers/igb/%s" % pci):
            self.igb_vfs[name] = 0
        else:
            if 0 == os.system('/sbin/dpdk-devbind --status-dev net | grep "unused=.*igb.*$" | grep "%s" -q' % pci):
                self.igb_vfs[name] = 0

def set_igb_param(self, name, mode):
    if name not in self.igb_vfs:
        return
    if mode == 'sr-iov':
        self.igb_vfs[name] = 7
    else:
        self.igb_vfs[name] = 0

    self.igb_recalc_needed = True

def recalc_igb_params(self):
    self.igb_recalc_needed = False
    max_vfs = None
    only_zeroes = True
    for n, vfs in self.igb_vfs.items():
        if max_vfs:
            max_vfs = "%s,%s" % (max_vfs, vfs)
        else:
            max_vfs = "%s" % (vfs)
        if vfs > 0:
            only_zeroes = False

    max_vfs_opt = "options igb max_vfs=%s" % max_vfs
    need_update = False
    try:
        optfile = open("/etc/modprobe.d/igb.conf", "r")
        cur_opt = optfile.read()
        optfile.close()
        if max_vfs_opt != cur_opt:
            if only_zeroes and bool(re.match('^options igb max_vfs=[0,]+$', cur_opt)):
                need_update = False
            else:
                need_update = True
    except:
        need_update = True
    try:
        if need_update:
            optfile = open("/etc/modprobe.d/igb.conf", 'w')
            optfile.write("options igb max_vfs=%s" % max_vfs)
            optfile.close()
            os.system('update-initramfs -u')
            self.initramfs_changed = True
    except:
        self.logger.error("Couldn't update \"/etc/modprobe.d/igb.conf\" file.")

def tag_interface(iface, vlan):
    if vlan is not None and vlan != 0:
        iface += '.' + str(vlan)
        return iface
    return iface

def vpp_not_running():
    if is_svc_masked('vpp-config.service'):
        return True
    if not os.path.isfile('/run/vpp.pid'):
        return True
    return False

def get_ssh_port(self):
    try:
        with open(SSH_PORT_FILE, 'r') as file:
            content = file.read().strip()
            ssh_port = int(content) if content else 22222

    except FileNotFoundError:
        self.logger.error(f"Error: The file {SSH_PORT_FILE} was not found. Using default port 22222.")
        ssh_port = 22222
    except Exception as e:
        self.logger.error(f"Error occurred: {e} while getting the ssh-port.")
        ssh_port = 22222

    return ssh_port

def str_get_ifname(str):
    r = re.compile(".*((data[0-9]{1,2})|(lte[0-9]{1,2})|(nat[0-9]{1,2})|(wifi0)|(mgmt[0-9]{1})).*")
    m = r.match(str)
    # self.logger.info('{} - {}'.format(str, None if m is None else m.groups()))
    if m is None:
        return None

    for name in m.groups():
        if name is not None:
            return name

    return None

def get_ifname_from_change(self, change_elem):
    name_id = change_elem['xpath'].lstrip(self.namespace).split('/', 1)
    # self.logger.info(name_id)
    if name_id:
        return str_get_ifname(name_id[0])

    return None

def do_ifup(self, ifname):
    if 'data' in ifname and 'br-' in ifname:
        kern_name = ifname.strip("br-")
        if os.path.exists("/sys/class/net/{}".format(kern_name)):
            foundup=os.system("grep -q up /sys/class/net/{}/operstate 1>/dev/null 2>/dev/null".format(kern_name))
            if foundup !=0:
                return
        else:
            founddown = os.system("/usr/bin/vppctl show hardware-interfaces {} | tr -d \"\r\" | head -5 | grep {} | grep -q down".format(kern_name, kern_name))
            if founddown==0:
                return

    shell_return_ok(self,"ifup --ignore-error {}".format(ifname))

def apply_management_rules(self, ifname, management_old, management, change_set):
    if not ifname:
        self.logger.debug('No ifname')
        return

    if 'true' in str(management_old).lower() or change_set == None:
        if 'false' in str(management).lower():
            ssh_port = get_ssh_port(self)
            shell_return_ok(self,"iptables -I INPUT 1 -i {} -p tcp -m tcp --dport {} -j REJECT --reject-with tcp-reset".format(ifname, ssh_port))
            shell_return_ok(self,"ip6tables -I INPUT 1 -i {} -p tcp -m tcp --dport {} -j REJECT --reject-with tcp-reset".format(ifname, ssh_port))
            shell_return_ok(self,"iptables -I INPUT 1 -i {} -p tcp -m tcp --dport 80 -j REJECT --reject-with tcp-reset".format(ifname))
            shell_return_ok(self,"ip6tables -I INPUT 1 -i {} -p tcp -m tcp --dport 80 -j REJECT --reject-with tcp-reset".format(ifname))
            shell_return_ok(self,"iptables -I INPUT 1 -i {} -p tcp -m tcp --dport 443 -j REJECT --reject-with tcp-reset".format(ifname))
            shell_return_ok(self,"ip6tables -I INPUT 1 -i {} -p tcp -m tcp --dport 443 -j REJECT --reject-with tcp-reset".format(ifname))
            shell_return_ok(self,"iptables -I INPUT 1 -i {} -p tcp -m tcp --dport 830 -j REJECT --reject-with tcp-reset".format(ifname))
            shell_return_ok(self,"ip6tables -I INPUT 1 -i {} -p tcp -m tcp --dport 830 -j REJECT --reject-with tcp-reset".format(ifname))
            shell_return_ok(self,"iptables -I INPUT 1 -i {} -p tcp -m tcp --dport 5000 -j REJECT --reject-with tcp-reset".format(ifname))
            shell_return_ok(self,"ip6tables -I INPUT 1 -i {} -p tcp -m tcp --dport 5000 -j REJECT --reject-with tcp-reset".format(ifname))
            shell_return_ok(self,"iptables -I INPUT 1 -i {} -p tcp -m tcp --dport 7373 -j REJECT --reject-with tcp-reset".format(ifname))
            shell_return_ok(self,"ip6tables -I INPUT 1 -i {} -p tcp -m tcp --dport 7373 -j REJECT --reject-with tcp-reset".format(ifname))
            shell_return_ok(self,"iptables -I INPUT 1 -i {} -p tcp -m tcp --dport 35357 -j REJECT --reject-with tcp-reset".format(ifname))
            shell_return_ok(self,"ip6tables -I INPUT 1 -i {} -p tcp -m tcp --dport 35357 -j REJECT --reject-with tcp-reset".format(ifname))

    if 'false' in str(management_old).lower():
        if 'true' in str(management).lower():
            ssh_port = get_ssh_port(self)
            shell_return_ok(self,"iptables -D INPUT -i {} -p tcp -m tcp --dport {} -j REJECT --reject-with tcp-reset".format(ifname, ssh_port))
            shell_return_ok(self,"ip6tables -D INPUT -i {} -p tcp -m tcp --dport {} -j REJECT --reject-with tcp-reset".format(ifname, ssh_port))
            shell_return_ok(self,"iptables -D INPUT -i {} -p tcp -m tcp --dport 80 -j REJECT --reject-with tcp-reset".format(ifname))
            shell_return_ok(self,"ip6tables -D INPUT -i {} -p tcp -m tcp --dport 80 -j REJECT --reject-with tcp-reset".format(ifname))
            shell_return_ok(self,"iptables -D INPUT -i {} -p tcp -m tcp --dport 443 -j REJECT --reject-with tcp-reset".format(ifname))
            shell_return_ok(self,"ip6tables -D INPUT -i {} -p tcp -m tcp --dport 443 -j REJECT --reject-with tcp-reset".format(ifname))
            shell_return_ok(self,"iptables -D INPUT -i {} -p tcp -m tcp --dport 830 -j REJECT --reject-with tcp-reset".format(ifname))
            shell_return_ok(self,"ip6tables -D INPUT -i {} -p tcp -m tcp --dport 830 -j REJECT --reject-with tcp-reset".format(ifname))
            shell_return_ok(self,"iptables -D INPUT -i {} -p tcp -m tcp --dport 5000 -j REJECT --reject-with tcp-reset".format(ifname))
            shell_return_ok(self,"ip6tables -D INPUT -i {} -p tcp -m tcp --dport 5000 -j REJECT --reject-with tcp-reset".format(ifname))
            shell_return_ok(self,"iptables -D INPUT -i {} -p tcp -m tcp --dport 7373 -j REJECT --reject-with tcp-reset".format(ifname))
            shell_return_ok(self,"ip6tables -D INPUT -i {} -p tcp -m tcp --dport 7373 -j REJECT --reject-with tcp-reset".format(ifname))
            shell_return_ok(self,"iptables -D INPUT -i {} -p tcp -m tcp --dport 35357 -j REJECT --reject-with tcp-reset".format(ifname))
            shell_return_ok(self,"ip6tables -D INPUT -i {} -p tcp -m tcp --dport 35357 -j REJECT --reject-with tcp-reset".format(ifname))

def config_smart_protocol_passthrough(self, name, config, change_set):
    if config is None:
        return

    ippass_enabled = config.get('ip-passthrough', False)
    if not ippass_enabled:
        return

    if change_set:
        change = next(iter(change_set.keys()))
        if 'smart-protocol-passthrough' in change:
            str_change_set = json.dumps(change_set)
            str_change_set = str_change_set[:-1] + f', "wan-interface": "{name}"}}'
            self.tracehost.on_smartpp_config_change(str_change_set, [], False)
            return

    # Startup
    if 'smart-protocol-passthrough' in config:
        startup_cfg = [json.dumps(cfg) for cfg in config['smart-protocol-passthrough']]
        startup_cfg = [protocol[:-1] + f', "wan-interface": "{name}"}}' for protocol in startup_cfg]
        if startup_cfg:
            self.tracehost.on_smartpp_config_change('', startup_cfg, True)

def restore_mac(interface, bridge, mode):
    if 'data' in bridge and not vpp_not_running():
        interface=str(interface.split('.')[0])
        if mode == 'virtio':
            cmd='A=$(cat /sys/class/net/%s/address); B=$(cat /sys/class/net/%s/address) ; if [ -f /sys/class/net/%s/address ] && [ -n "$B" ] && [ -n "$A" ] && [ "$A" != "$B" ] ; then ifconfig %s hw ether $A; logger "Restoring MAC of %s";  fi' % (interface, bridge, bridge, bridge, bridge )
        else:
            cmd='A=$(vppctl show hard %s | grep Ethern | xargs | cut -f3 -d" "); B=$(cat /sys/class/net/%s/address) ; if [ -f /sys/class/net/%s/address ] && [ -n "$B" ] && [ -n "$A" ] && [ "$A" != "$B" ] ; then ifconfig %s hw ether $A; logger "Restoring MAC of %s";  fi' % (interface, bridge, bridge, bridge, bridge )

        os.system(cmd)
