#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import ipaddress
import re
import subprocess
import vppapi_if.vppapi_if as vppif
from sysrepo_interfaces import utils as utils
import ucpeif.vxlanif as vxlif

def set_ippass_enable_vxlan(self, ifname, lvtep, rvteps, tun_type, enable):
    if utils.vpp_not_running():
        return

    vpp_if = vppif.VppApiInterface('sr-interfaces')
    mode = self.get_item("/ucpe:config/interfaces:interface[name='{}']/mode".format(ifname))
    if mode == 'virtio':
        if_prefix = 'host-'
    else:
        if_prefix = ''

    vpp_idx = vpp_if.get_ifidx_by_name('{}{}'.format(if_prefix, ifname))
    if not vpp_idx:
        del(vpp_if)
        return

    ippass_host_idx = vpp_if.ippass_get_host_if_idx(vpp_idx)
    for rvtep in rvteps:
        if type(rvtep) != type(lvtep):
            continue
        ret = vpp_if.add_del_ippass_permit_host_entry(host_idx=ippass_host_idx,
                                                      protocol='udp',
                                                      dir_out=0,
                                                      dst_port=4790 if tun_type == 'vxlan-gpe' else 4789,
                                                      ip=lvtep,
                                                      src_ip=rvtep,
                                                      add=1 if enable else 0)
        self.logger.debug('add_del_ippass_permit_host_entry ret {})'.format(ret))

    del(vpp_if)

def get_vxlan_lvtep_ippass(self, lvtep):
    if_cfg = self.get_interfaces()
    ippass_ifs = list(filter(lambda x: True if (x.get('type', '') in ['ethernet', 'lte', 'wifi', 'pppoe']) and x.get('ip-passthrough', False) else False, list(if_cfg.values())))
    for interface in ippass_ifs:
        if lvtep.version == 6:
            address = interface.get('ipv6-prefix', None)
        else:
            address = interface.get('ipv4-prefix', None)
        if address is not None and ipaddress.ip_interface(address).ip == lvtep:
            return (interface.get('name', None), True)

    return (None, False)

def config_ippass_vxlan(self, old_local, old_remote, old_type, new_local=None, new_remote=[], new_type=None, startup=False):
    """ Have to find the outgoing interface from the local VTEP address confguration
            and to check if ip-passthrough is enabled.
        Then according to the Local + Remotes to add permit entry in ip-pass config to match:
            IPv4 / IPv6 UPD packets
            Dst ip - local VTEP
            Src ip - Remote VTEP (for every remote address in list)
            Dst port 4789 for vxlan ot 4790 for vxlan-gpe
    """
    if not old_remote and not new_remote and old_local == None and new_local == None:
        return
    self.logger.debug('config_ippass_vxlan({},{},{},{},{},{},{})'.format(old_local, old_remote, old_type, new_local, new_remote, new_type, startup))

    old_ippass_enabled = False
    new_ippass_enabled = False
    old_ifname = None
    new_ifname = None

    if old_local:
        (old_ifname, old_ippass_enabled) = get_vxlan_lvtep_ippass(self, old_local)
    if new_local:
        (new_ifname, new_ippass_enabled) = get_vxlan_lvtep_ippass(self, new_local)

    if startup == True:
        if old_ippass_enabled and old_local and old_remote and old_type:
            self.logger.debug('VxLAN startup config - {} {} {} {}'.format(old_ifname, old_local, old_remote, old_type))
            set_ippass_enable_vxlan(self, old_ifname, old_local, old_remote, old_type, True)
    else:
        if old_ippass_enabled and old_local and old_remote and old_type:
            self.logger.debug('VxLAN running remove config - {} {} {} {}'.format(old_ifname, old_local, old_remote, old_type))
            set_ippass_enable_vxlan(self, old_ifname, old_local, old_remote, old_type, False)

        if new_ippass_enabled or old_ippass_enabled:
            ifname = None
            ifname = new_ifname if new_ippass_enabled else old_ifname

            if new_local and new_remote and new_type:
                self.logger.debug('VxLAN running apply config - {} {} {} {}'.format(ifname, new_local, new_remote, new_type))
                set_ippass_enable_vxlan(self, ifname, new_local, new_remote, new_type, True)

def config_interface_vxlan(self, name, config, change_set):
    if self.get_network_mode() != 'vpp':
        self.logger.error('uCPE is not in VPP mode')
        return
    self.logger.debug("config_interface_vxlan")
    self.logger.debug("Config {}".format(config))
    self.logger.debug("Change set {}".format(change_set))

    local = None
    lvtepv4 = None
    lvtepv6 = None
    old_local = None
    old_localv6 = None
    old_remote = []
    old_type = None
    lvtep = None
    rvteps_add = []
    rvteps_del = []
    tun_type = None
    mcast = None
    mcast_ip = None
    old_mcast_ip = None

    if config:
        if config.get('local', {}).get('vtep-addr', None) != None:
            local = config.get('local', {}).get('vtep-addr', None)
            if local:
                old_local = ipaddress.ip_address(local) if isinstance(ipaddress.ip_address(local), ipaddress.IPv4Address) else None
                old_localv6 = ipaddress.ip_address(local) if isinstance(ipaddress.ip_address(local), ipaddress.IPv6Address) else None
        elif config.get('local', {}).get('vtep-if', None) != None:
            local = config.get('local').get('vtep-if', None)
            if local != None:
                bootproto = None
                address = None
                addressv6 = None
                bootproto = self.get_item("/ucpe:config/interfaces:interface[name='{}']/bootproto".format(local))
                if bootproto == "none" and bootproto != None:
                    addressv6 = self.get_item("/ucpe:config/interfaces:interface[name='{}']/ipv6-prefix".format(local))
                    if addressv6 is not None:
                        old_localv6 = ipaddress.ip_interface(addressv6).ip

                    address = self.get_item("/ucpe:config/interfaces:interface[name='{}']/ipv4-prefix".format(local))
                    if address is not None:
                        old_local = ipaddress.ip_interface(address).ip

        for rvtep_addr in config.get('remote', {}).get('vtep-addrs', []):
            old_remote.append(ipaddress.ip_address(rvtep_addr.get('vtep-addr')))

        old_type = config.get('tunnel-type', 'vxlan')

        if config.get('remote', None) != None:
            mcast = config.get('remote').get('mcast-vtep-addr', None)
            if mcast != None:
                old_mcast_ip = ipaddress.ip_address(mcast)

    if not change_set and config:
        # Startup config
        vxlan_if = next((a for a in self.vxlan_interfaces if a.name == config.get('name')), None)
        tun_type = vxlif.TUNNEL_TYPE.VXLAN
        if vxlan_if:
            self.logger.debug("Interface \'{}\' allready exists.".format(vxlan_if.name))
        else:
            vni = config.get('vni', None)
            shg = config.get('shg', None)

            if old_type == 'vxlan-gpe':
                tun_type = vxlif.TUNNEL_TYPE.VXLAN_GPE

            vxlan_if = vxlif.VxLanInterface( config.get('name'), vni=vni, lvtep=old_local, lvtepv6=old_localv6, rvteps=old_remote, tun_type=tun_type,
                                             logger=self.logger, shg=shg, mcast_if=local, mcast_ip=old_mcast_ip)
            self.vxlan_interfaces.append(vxlan_if)
            enable = config.get('enable', False)
            vxlan_if.enable = enable

            ipv4_pr = config.get('ipv4-prefix', None)
            if ipv4_pr:
                vxlan_if.ip4 = ipaddress.IPv4Interface(str(ipv4_pr))

            ipv6_pr = config.get('ipv6-prefix', None)
            if ipv6_pr:
                vxlan_if.ip6 = ipaddress.IPv6Interface(str(ipv6_pr))

            bootproto = config.get('bootproto', None)
            bp_set = None
            if bootproto:
                if bootproto == 'none':
                    bp_set = vxlif.BOOT_PROTO.NONE
                elif bootproto == 'dhcp':
                    bp_set = vxlif.BOOT_PROTO.DHCP
                elif bootproto == 'dhcp6':
                    bp_set = vxlif.BOOT_PROTO.DHCP6
                elif bootproto == 'dhcp46':
                    bp_set = vxlif.BOOT_PROTO.DHCP46

                if bp_set:
                    vxlan_if.bootproto = bp_set

            dns = config.get('dns', None)
            if dns:
                self.logger.debug('! Setting DNS with name {}'.format(name))
                dnses = dns.split(',')
                dnsopts = ''.join(map(lambda dns_srv: ' '+ dns_srv, dnses))
                subprocess.call('/usr/bin/resolvectl dns br-{} {}'.format(name, dnsopts), shell=True)

        if old_local:
            config_ippass_vxlan(self, None, [], None, old_local, old_remote, old_type)
        if old_localv6:
            config_ippass_vxlan(self, None, [], None, old_localv6, old_remote, old_type)
        return

    if change_set:
        ch_set_keys = change_set.keys()
        lvtep_set = False
        rvteps_set = False
        tuntype_set = False

        self.logger.info('Keys : {}'.format(ch_set_keys))
        vxlan_if = next((vxlan for vxlan in self.vxlan_interfaces if vxlan.name == name), None)

        if 'name' in ch_set_keys:
            if change_set.get('name', None):
                if change_set.get('name').get('operation', None) == 'CREATED':
                    if vxlan_if:
                        self.logger.error("There is object for {} ?!?!".format(name))
                    else:
                        vxlan_if = vxlif.VxLanInterface(name, logger=self.logger)
                        self.vxlan_interfaces.append(vxlan_if)
                elif change_set.get('name').get('operation', None) == 'DELETED':
                    if vxlan_if:
                        self.vxlan_interfaces.remove(vxlan_if)
                        del vxlan_if
                    if old_local:
                        config_ippass_vxlan(self, old_local, old_remote, old_type, None, [], None)
                    if old_localv6:
                        config_ippass_vxlan(self, old_localv6, old_remote, old_type, None, [], None)
                    return

        if 'vni' in ch_set_keys:
            self.logger.debug('VNI is changed')

            oper = change_set.get('vni').get('operation', None)
            vni = None

            if oper == 'MODIFIED' or oper == 'CREATED':
                vni = change_set.get('vni').get('new_value', None)

            vxlan_if.vni = int(vni)

            self.logger.info('VNI operation is {}'.format(oper))

        if 'shg' in ch_set_keys:
            self.logger.debug('SHG is changed')

            oper = change_set.get('shg').get('operation', None)
            shg = 100

            if oper == 'MODIFIED' or oper == 'CREATED':
                shg = change_set.get('shg').get('new_value', None)

            vxlan_if.shg = int(shg)

            self.logger.info('SHG operation is {}'.format(oper))

        if 'local/vtep-addr' in ch_set_keys:
            self.logger.debug('local/vtep-addr is changed')
            oper = change_set.get('local/vtep-addr').get('operation', None)
            lvtep_set = True
            lvtep = None
            if oper == 'CREATED' or oper == 'MODIFIED':
                lvtep = change_set.get('local/vtep-addr').get('new_value', None)
            elif oper == 'DELETED':
                if 'local/vtep-if' not in ch_set_keys:
                    self.logger.debug('local/vtep-addr is deleted')
                    vxlan_if.lvtep = None

            if lvtep:
                lvtep = ipaddress.ip_address(lvtep)
                if lvtep.version == 6:
                    if old_local:
                        vxlan_if.lvtep = None

                    vxlan_if.lvtepv6 = lvtep
                    lvtepv6 = lvtep
                else:
                    if old_localv6:
                        vxlan_if.lvtepv6 = None

                    vxlan_if.lvtep = lvtep
                    lvtepv4 = lvtep

            self.logger.debug('LOCAL ip operation is {}'.format(oper))

        if 'local/vtep-if' in ch_set_keys:
            self.logger.debug('local/vtep-if is changed')
            oper = change_set.get('local/vtep-if').get('operation', None)
            lvtep_set = True
            lvtep = None
            if oper == 'CREATED' or oper == 'MODIFIED':
                lvtep = change_set.get('local/vtep-if').get('new_value', None)
                vxlan_if.mcast_if = lvtep
            elif oper == 'DELETED':
                if 'local/vtep-addr' not in ch_set_keys:
                    self.logger.debug('local/vtep-if is deleted')
                    vxlan_if.lvtep = None
                    vxlan_if.lvtepv6 = None

                vxlan_if.mcast_if = None

            if lvtep:
                bootproto = None
                address = None
                addressv6 = None
                bootproto = self.get_item("/ucpe:config/interfaces:interface[name='{}']/bootproto".format(lvtep))
                if (bootproto == "none" and bootproto != None) or "lo" in lvtep:
                    addressv6=self.get_item("/ucpe:config/interfaces:interface[name='{}']/ipv6-prefix".format(lvtep))
                    if addressv6 is not None:
                        lvtepv6 = ipaddress.ip_interface(addressv6).ip

                    address = self.get_item("/ucpe:config/interfaces:interface[name='{}']/ipv4-prefix".format(lvtep))
                    if address is None:
                        address = change_set.get('local/lvtep-if-ipv4-prefix', None).get('new_value', None)
                    if address is not None:
                        lvtepv4 = ipaddress.ip_interface(address).ip
                else:
                    lvtepv4 = None

                if lvtepv4:
                    vxlan_if.lvtep = lvtepv4
                if lvtepv6:
                    vxlan_if.lvtepv6 = lvtepv6

            self.logger.debug('LOCAL if operation is {}'.format(oper))

        p = re.compile("^remote\/vtep-addrs\[vtep-addr='.*'\]\/vtep-addr")

        for key in ch_set_keys:
            if p.match(key):
                oper = change_set.get(key).get('operation', None)
                rvteps_set = True
                if oper == 'CREATED':
                    rvteps_add.append(ipaddress.ip_address(change_set.get(key).get('new_value')))
                elif oper == 'DELETED':
                    rvteps_del.append(ipaddress.ip_address(change_set.get(key).get('old_value')))
                else:
                    self.logger.info("Unhandled operation for REMOTE {}".format(oper))

                self.logger.info('REMOTE operation is {}'.format(oper))

        if rvteps_add:
            vxlan_if.add_rvteps(rvteps_add)

        if rvteps_del:
            vxlan_if.del_rvteps(rvteps_del)

        if 'remote/mcast-vtep-addr' in ch_set_keys:
            self.logger.debug('remote/mcast-vtep-addr is changed')
            oper = change_set.get('remote/mcast-vtep-addr').get('operation', None)
            mcast = None
            if oper == 'CREATED' or oper == 'MODIFIED':
                mcast = change_set.get('remote/mcast-vtep-addr').get('new_value', None)
                vxlan_if.mcast_ip = ipaddress.ip_address(mcast)
            elif oper == 'DELETED':
                self.logger.debug('remote/mcast-vtep-addr is deleted')
                vxlan_if.mcast_ip = None

            # if lvtep:
            #     lvtep = ipaddress.ip_address(lvtep)
            #     vxlan_if.lvtep = lvtep

        if 'enable' in ch_set_keys:
            oper = change_set.get('enable').get('operation', None)
            if oper == 'MODIFIED' or oper == 'CREATED':
                enable = change_set.get('enable').get('new_value')
                enable = True if 'true' in str(enable).lower() else False
            else:
                enable = False

            vxlan_if.enable = enable

        if 'ipv4-prefix' in ch_set_keys:
            oper = change_set.get('ipv4-prefix').get('operation', None)
            if oper == 'MODIFIED' or oper == 'CREATED':
                ipv4_pr = change_set.get('ipv4-prefix').get('new_value')
                vxlan_if.ip4 = ipaddress.IPv4Interface(str(ipv4_pr))
            else:
                vxlan_if.ip4 = None

        if 'ipv6-prefix' in ch_set_keys:
            oper = change_set.get('ipv6-prefix').get('operation', None)
            if oper == 'MODIFIED' or oper == 'CREATED':
                ipv6_pr = change_set.get('ipv6-prefix').get('new_value')
                vxlan_if.ip6 = ipaddress.IPv6Interface(str(ipv6_pr))
            else:
                vxlan_if.ip6 = None

        if 'bootproto' in ch_set_keys:
            oper = change_set.get('bootproto').get('operation', None)
            bp_set = None
            if oper == 'MODIFIED' or oper == 'CREATED':
                bootproto = change_set.get('bootproto').get('new_value')

                if bootproto == 'none':
                    bp_set = vxlif.BOOT_PROTO.NONE
                elif bootproto == 'dhcp':
                    bp_set = vxlif.BOOT_PROTO.DHCP
                elif bootproto == 'dhcp6':
                    bp_set = vxlif.BOOT_PROTO.DHCP6
                elif bootproto == 'dhcp46':
                    bp_set = vxlif.BOOT_PROTO.DHCP46

                if bp_set:
                    vxlan_if.bootproto = bp_set
            else:
                vxlan_if.bootproto = vxlif.BOOT_PROTO.NONE

        if 'tunnel-type' in ch_set_keys:
            oper = change_set.get('tunnel-type').get('operation', None)
            tuntype_set = True
            type_set = None
            if oper == 'MODIFIED' or oper == 'CREATED':
                tun_type = change_set.get('tunnel-type').get('new_value')

                if tun_type == 'vxlan':
                    type_set = vxlif.TUNNEL_TYPE.VXLAN
                elif tun_type == 'vxlan-gpe':
                    type_set = vxlif.TUNNEL_TYPE.VXLAN_GPE

                if type_set:
                    vxlan_if.tun_type = type_set
            else:
                vxlan_if.tun_type = vxlif.TUNNEL_TYPE.VXLAN

        if 'dns' in ch_set_keys:
            oper = change_set.get('dns').get('operation', None)
            if oper == 'MODIFIED' or oper == 'CREATED':
                dns = change_set.get('dns').get('new_value')
                dnses = dns.split(',')
                dnsopts = ''.join(map(lambda dns_srv: ' '+ dns_srv, dnses))
                subprocess.call('/usr/bin/resolvectl dns br-{} {}'.format(name, dnsopts), shell=True)
            else:
                dns = change_set.get('dns').get('old_value')
                dnses = dns.split(',')
                # revert all !?
                subprocess.call('/usr/bin/resolvectl revert br-{}'.format(name), shell=True)

        for key in change_set.keys():
            self.logger.debug('Key \'{}\' operation {}'.format(key, change_set.get(key).get('operation')))

        if lvtep_set or rvteps_set or tuntype_set:
            if not rvteps_set:
                rvteps_del = old_remote
                rvteps_add = old_remote
            if not lvtep_set:
                lvtepv4 = old_local
                lvtepv6 = old_localv6
            if not tuntype_set:
                tun_type = old_type

            config_ippass_vxlan(self, old_local, rvteps_del, old_type,
                                lvtepv4, rvteps_add, tun_type, False)
            config_ippass_vxlan(self, old_localv6, rvteps_del, old_type,
                                lvtepv6, rvteps_add, tun_type, False)
