#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import ipaddress
from sysrepo_interfaces import utils as utils
import ucpeif.wireguardvpnif as wireguardif
import vppapi_if.vppapi_if as vppif
import socket

def add_del_ip_pass_tap_intf_if_wg_enabled(self, ifname, add_del):
    if_cfg = self.get_interfaces()
    ippass_ifs = list(filter(lambda x: True if (x.get('type', '') in ['ethernet', 'lte', 'vlan', 'wifi', 'pppoe']) and
        x.get('ip-passthrough', False) else False, list(if_cfg.values())))

    local_endpoint = self.get_interface_param(ifname, 'local-endpoint', '')

    vpp_if = vppif.VppApiInterface('sr-interfaces')
    for interface in ippass_ifs:
        intfname = interface.get('name', 'noname')
        ippass_prefix = self.get_ip_mask_kernel_intfc(f'br-{intfname}')

        if local_endpoint == ippass_prefix[0]:
            local_port = self.get_interface_param(ifname, 'local-port', '')
            peer_endpoint = self.get_wireguard_peer_params(ifname, 'peer-endpoint')

            if local_endpoint and local_port:

                (if_name, vpp_idx) = vpp_if.search_vppif_by_hostifname(intfname)
                ippass_host_idx = vpp_if.ippass_get_host_if_idx(vpp_idx)
                bridge_data = vpp_if.get_interface_bd(ippass_host_idx)

                sw_if_index_values = []
                for sw_if_detail in bridge_data['sw_if_details']:
                    sw_if_index_values.append(sw_if_detail['sw_if_index'])

                for idx in sw_if_index_values:
                    intf_name = vpp_if.get_ifname_by_ifidx(idx)
                    if intf_name.startswith('tap'):
                        vpp_if.add_del_ippass_permit_host_entry(host_idx=idx, protocol='tcp', dst_port=7374)
                        vpp_if.add_del_ippass_permit_host_entry(host_idx=idx, protocol='tcp', dst_port=7373)

                        if peer_endpoint:
                            if self.get_interface_param(ifname, 'enable') == False or not self.get_interface_param(ifname, 'enable'):
                                ret = vpp_if.add_del_ippass_permit_host_entry(idx, 'udp', 0, int(local_port), local_endpoint,
                                                                        peer_endpoint, 0)
                                break
                            else:
                                ret = vpp_if.add_del_ippass_permit_host_entry(idx, 'udp', 0, int(local_port), local_endpoint,
                                                                            peer_endpoint, add_del)
                        break
    del vpp_if

def config_interface_wireguard(self, name, config, change_set):
    if self.get_network_mode() != 'vpp':
        self.logger.error('uCPE is not in VPP mode')
        return

    c_enabled = False
    c_l_port = None
    c_l_endpoint = None
    c_management = False
    c_local = None
    c_localv6 = None
    n_local = None
    n_localv6 = None

    self.logger.debug("Config {}".format(config))
    if config:        
        c_enabled = config.get('enable', False)
        c_management = config.get('management', False)
        c_l_port = config.get('local-port', None)
        ipv4_pr = config.get('ipv4-prefix', None)
        ipv6_pr = config.get('ipv6-prefix', None)

        ifname = utils.get_ifname(self,name)

        if config.get('local-endpoint', None):
            c_l_endpoint = config.get('local-endpoint', None)
            c_local, c_localv6 = get_local_wireguard_endpoint_ip(self, c_l_endpoint)
        
        concentrator_endpoint = get_wireguard_concentrator_param(config, 'concentrator-endpoint')
        c_endpoint = get_ip_from_hostname(self, concentrator_endpoint)
        concentrator_password = get_wireguard_concentrator_param(config, 'concentrator-password')
        concentrator_port = get_wireguard_concentrator_param(config, 'concentrator-port')
        domain_lookup = get_wireguard_concentrator_param(config, 'domain-lookup')

        self.logger.debug(f"{c_endpoint} {concentrator_password} {concentrator_port} {domain_lookup}")
        
    if not change_set and config:
        # Startup config
        self.logger.debug("APPLYING STARTUP FOR WIREGUARD INTERFACE - {}".format(name))
        wg_if = next(( a for a in self.wg_interfaces if a.name == config.get('name')), None)
        if wg_if:
            self.logger.debug("Interface \'{}\' allready exists.".format(wg_if.name))
        else:
            self.logger.debug("Add WireGuard configuration for \'{}\' interface" .format(config.get('name')))
            wg_if = wireguardif.WireGuardVPNInterface(config.get('name'), logger=self.logger, l_port=c_l_port,
                                                      l_endpoint=c_local, l_endpoint_v6=c_localv6, conc_endpoint=c_endpoint,
                                                      conc_password=concentrator_password, conc_port=concentrator_port,
                                                      enable=c_enabled, d_lookup=domain_lookup, wg_count=len(self.wg_interfaces))

            if 'peers' in config:
                for peer in config['peers']:
                    self.logger.debug("Peer {}".format(peer))
                    r_endpoint = peer.get('peer-endpoint', None)
                    if not r_endpoint:
                        continue
                    r_port = peer.get('peer-port', None)
                    r_ips = peer.get('allowed-ips', None)
                    r_keepalive = peer.get('persistent-keepalive', 0)
                    wg_if.init_peer(r_endpoint=r_endpoint, r_port=r_port, allowed_ips=r_ips, persistent_keepalive=r_keepalive)

            self.wg_interfaces.append(wg_if)

            mtu = config.get('mtu', 1420)
            if mtu:
                wg_if.mtu = int(mtu)

            if ipv4_pr:
                wg_if.ip4_addr = ipaddress.IPv4Interface(str(ipv4_pr))

            if ipv6_pr:
                wg_if.ip6_addr = ipaddress.IPv6Interface(str(ipv6_pr))

            add_del_ip_pass_tap_intf_if_wg_enabled(self, ifname, 1)
            utils.apply_management_rules(self, ifname, False, c_management, change_set)

    if change_set:
        self.logger.debug("APPLYING RUNNING FOR WIREGUARD INTERFACE - {}".format(name))
        # self.logger.debug('change_set {} ' .format(change_set))
        ch_set_keys = change_set.keys()
        self.logger.debug('Keys : {}'.format(ch_set_keys))

        wg_if = next(( a for a in self.wg_interfaces if a.name == name), None)

        if 'name' in ch_set_keys:
            if change_set.get('name', None):
                if change_set.get('name').get('operation', None) == 'CREATED':
                    if wg_if:
                        self.logger.error("There is WireGuard interface object for {} ?!?!".format(name))
                    else:
                        wg_if = wireguardif.WireGuardVPNInterface( name, logger=self.logger, wg_count=len(self.wg_interfaces) )
                        self.wg_interfaces.append(wg_if)
                elif change_set.get('name').get('operation', None) == 'DELETED':
                    if wg_if:
                        self.wg_interfaces.remove(wg_if)
                        del wg_if
                    return

        if 'local-port' in ch_set_keys:
            oper = change_set.get('local-port').get('operation', None )
            n_l_port = None

            self.logger.debug('Local port is changed - operation {}' .format(oper))
            if oper == 'MODIFIED' or oper == 'CREATED':
                n_l_port = change_set.get('local-port').get('new_value', None)

            wg_if.l_port = int(n_l_port)

        if 'local-endpoint' in ch_set_keys:
            oper = change_set.get('local-endpoint').get('operation', None )
            n_l_endpoint = None

            self.logger.debug('Local endpoint is changed - operation {}' .format(oper))
            if oper == 'MODIFIED' or oper == 'CREATED':
                n_l_endpoint = change_set.get('local-endpoint').get('new_value', None)

            n_local, n_localv6 = get_local_wireguard_endpoint_ip(self, n_l_endpoint)
            
            if n_local:
                wg_if.l_endpoint = n_local
                self.logger.debug('Local endpoint configured {}' .format(n_local))
            if n_localv6:
                wg_if.l_endpoint_v6 = n_localv6
                self.logger.debug('Local endpoint configured {}' .format(n_local))

        if 'mtu' in ch_set_keys:
            oper = change_set.get('mtu').get('operation', None )
            n_mtu = 1420

            self.logger.debug('MTU is changed - operation {}' .format(oper))
            if oper == 'MODIFIED' or oper == 'CREATED':
                n_mtu = change_set.get('mtu').get('new_value', 1420)

            wg_if.mtu = int(n_mtu)

        if 'enable' in ch_set_keys:
            oper = change_set.get('enable').get('operation', None)
            n_enable = False

            self.logger.debug('State is changed - operation {}' .format(oper))
            if oper == 'MODIFIED' or oper == 'CREATED':
                if 'true' in str(change_set.get('enable').get('new_value', False)).lower():
                    n_enable = True
                else:
                    add_del_ip_pass_tap_intf_if_wg_enabled(self, ifname, 0)
                
            wg_if.enable = n_enable
        
        if 'ipv4-prefix' in ch_set_keys:
            oper = change_set.get('ipv4-prefix').get('operation', None)
            n_ip4_addr = None

            self.logger.debug('IPv4 prefix is changed - operation {}' .format(oper))
            if oper == 'MODIFIED' or oper == 'CREATED':
                n_ip4_addr = change_set.get('ipv4-prefix').get('new_value', None)
                if n_ip4_addr:
                    n_ip4_addr = ipaddress.IPv4Interface(str(n_ip4_addr))
                
            wg_if.ip4_addr = n_ip4_addr

        if 'ipv6-prefix' in ch_set_keys:
            oper = change_set.get('ipv6-prefix').get('operation', None)
            n_ip6_addr = None

            self.logger.debug('IPv4 prefix is changed - operation {}' .format(oper))
            if oper == 'MODIFIED' or oper == 'CREATED':
                n_ip6_addr = change_set.get('ipv6-prefix').get('new_value', None)
                if n_ip6_addr:
                    n_ip6_addr = ipaddress.IPv6Interface(str(n_ip6_addr))
                
            wg_if.ip6_addr = n_ip6_addr

        if 'management' in ch_set_keys:
            oper = change_set.get('management').get('operation', None)
            n_management = False

            self.logger.debug('Management is changed - operation {}' .format(oper))
            if oper == 'MODIFIED' or oper == 'CREATED':
                if 'true' in str(change_set.get('management').get('new_value', False)).lower():
                    n_management = True

            utils.apply_management_rules(self, name, c_management, n_management, change_set)
            
        if 'wireguard-concentrator/domain-lookup' in ch_set_keys:
            oper = change_set.get('wireguard-concentrator/domain-lookup').get('operation', None )
            d_l_lookup = False

            self.logger.error('Domain lookup is changed - operation {}' .format(oper))
            if oper == 'MODIFIED' or oper == 'CREATED':
                if 'true' in str(change_set.get('wireguard-concentrator/domain-lookup').get('new_value', False)).lower():
                    self.logger.debug(f"DOMAIN LOOKUP TRUE {change_set.get('wireguard-concentrator/domain-lookup')}")
                # d_l_lookup = change_set.get('wireguard-concentrator/domain-lookup').get('new_value', None)
                    d_l_lookup = True

            wg_if.d_lookup = d_l_lookup

        if 'wireguard-concentrator/concentrator-endpoint' in ch_set_keys:
            oper = change_set.get('wireguard-concentrator/concentrator-endpoint').get('operation', None )
            c_endpoint = None

            self.logger.error('Concentrator endpoint is changed - operation {}' .format(oper))
            if oper == 'MODIFIED' or oper == 'CREATED':
                c_endpoint = change_set.get('wireguard-concentrator/concentrator-endpoint').get('new_value', None)

            c_endpoint = get_ip_from_hostname(self, c_endpoint)
            c_local, c_localv6 = get_local_wireguard_endpoint_ip(self, c_endpoint)
            
            if c_local:
                wg_if.conc_endpoint = c_local
                self.logger.error('Concentrator endpoint configured {}'.format(c_local))
 
        if 'wireguard-concentrator/concentrator-port' in ch_set_keys:
            oper = change_set.get('wireguard-concentrator/concentrator-port').get('operation', None )
            c_port = None

            self.logger.error('Concentrator port is changed - operation {}' .format(oper))
            if oper == 'MODIFIED' or oper == 'CREATED':
                c_port = change_set.get('wireguard-concentrator/concentrator-port').get('new_value', None)

            wg_if.conc_port = int(c_port)
            
        if 'wireguard-concentrator/concentrator-password' in ch_set_keys:
            oper = change_set.get('wireguard-concentrator/concentrator-password').get('operation', None )
            c_pass = None

            self.logger.error('Concentrator port is changed - operation {}' .format(oper))
            if oper == 'MODIFIED' or oper == 'CREATED':
                c_pass = change_set.get('wireguard-concentrator/concentrator-password').get('new_value', None)

            wg_if.conc_password = c_pass

        if str(change_set).find('peers') > -1:
            for xpath, change in change_set.items():
                self.logger.debug('Peer xpath {} change {}' .format(xpath, change))
                if "peers" not in xpath:
                    continue
                # Get peer endpoint key
                peer_key = change['xpath'][change['xpath'].find('peers[peer-endpoint=\'')+len('peers[peer-endpoint=\''):change['xpath'].rfind('\'')]
                self.logger.debug('Peer config is changed for peer {}' .format(peer_key))
            
                if xpath.find('/peer-endpoint') > -1:
                    # Peer created or deleted
                    oper = change['operation']
                    self.logger.debug('Peer peer-endpoint config is changed - operation {}' .format(oper))
                    if oper == 'MODIFIED' or oper == 'CREATED':
                        wg_if.add_peer_param(peer_key)
                    elif oper == 'DELETED':
                        wg_if.del_peer_param(peer_key)

                if xpath.find('/peer-port') > -1:
                    oper = change['operation']
                    n_r_port = None
                    self.logger.debug('Peer peer-port config is changed - operation {}' .format(oper))
                    if oper == 'MODIFIED' or oper == 'CREATED':
                        n_r_port = change.get('new_value', None)
                    
                    wg_if.add_peer_param(peer_key, "r_port", n_r_port)

                if xpath.find('/allowed-ips') > -1:
                    oper = change['operation']
                    n_r_nets = None
                    self.logger.debug('Peer allowed-ips config is changed - operation {}' .format(oper))
                    if oper == 'MODIFIED' or oper == 'CREATED':
                        n_r_nets = change.get('new_value', None)
                    
                    wg_if.add_peer_param(peer_key, "allowed_ips", n_r_nets)

                if xpath.find('/persistent-keepalive') > -1:
                    oper = change['operation']
                    n_r_keepalive = None
                    self.logger.debug('Peer persistent-keepalive config is changed - operation {}' .format(oper))
                    if oper == 'MODIFIED' or oper == 'CREATED':
                        n_r_keepalive = change.get('new_value', None)
                    
                    wg_if.add_peer_param(peer_key, "persistent_keepalive", n_r_keepalive)

        add_del_ip_pass_tap_intf_if_wg_enabled(self, name, 1)

def get_wireguard_concentrator_param(config, param):
    return config.get('wireguard-concentrator', {}).get(param, '')

def get_local_wireguard_endpoint_ip(self, endpoint):
    if 'data' in endpoint or 'lte' in endpoint or 'wifi' in endpoint:
        bootproto = None
        address = None
        addressv6 = None
        bootproto = self.get_item("/ucpe:config/interfaces:interface[name='{}']/bootproto".format(endpoint))
        if bootproto == "none" and bootproto != None:
            addressv6 = self.get_item("/ucpe:config/interfaces:interface[name='{}']/ipv6-prefix".format(endpoint))
            if addressv6 is not None:
                n_localv6 = ipaddress.ip_interface(addressv6).ip
            else:
                n_localv6 = None
            address = self.get_item("/ucpe:config/interfaces:interface[name='{}']/ipv4-prefix".format(endpoint))
            if address is not None:
                n_local = ipaddress.ip_interface(address).ip
    elif endpoint != None and endpoint != "":
        n_local = ipaddress.ip_address(endpoint) if isinstance(ipaddress.ip_address(endpoint), ipaddress.IPv4Address) else None
        n_localv6 = ipaddress.ip_address(endpoint) if isinstance(ipaddress.ip_address(endpoint), ipaddress.IPv6Address) else None

    return (n_local, n_localv6)

def get_ip_from_hostname(self, conc_endpoint):
    try:
        ip = socket.gethostbyname(conc_endpoint)
        return ip
    except socket.gaierror as e:
        self.logger.error(f'Invalid hostname, error raised is {e}')

def del_iptables_reject_rule_7373(self, intf):
    try:
        subprocess.check_output(
            ['/sbin/iptables', '-D', 'INPUT', '-i', intf, '-p', 'tcp', '-m', 'tcp', '--dport', '7373', '-j', 'REJECT', '--reject-with', 'tcp-reset'], stderr=subprocess.STDOUT)
    except Exception as e:
        self.logger.error(f"Iptables rule for 7373 does not exist! Exception: {e}")
